﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2013 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Linq;
using HeuristicLab.Analysis.SolutionCaching;
using HeuristicLab.Analysis.SolutionCaching.RealVectorEncoding;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Analysis.AlgorithmBehavior.Analyzers {
  [StorableClass]
  [Item("RealVectorConvexHullSolutionCacheAnalyzer", "Calculates convex hull-based measures from real vector solution caches.")]
  public class RealVectorConvexHullSolutionCacheAnalyzer : SolutionCacheAnalyzer<RealVector, RealVectorSolutionInformation> {
    [StorableConstructor]
    private RealVectorConvexHullSolutionCacheAnalyzer(bool deserializing) : base(deserializing) { }
    private RealVectorConvexHullSolutionCacheAnalyzer(RealVectorConvexHullSolutionCacheAnalyzer original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new RealVectorConvexHullSolutionCacheAnalyzer(this, cloner);
    }

    public RealVectorConvexHullSolutionCacheAnalyzer() : base() { }

    public override void ExecuteAnalysis(ResultCollection results, SolutionCache<RealVector, RealVectorSolutionInformation> solutionCache) {
      int i = 0;
      var sols = solutionCache.GetSolutionsFromGeneration(i);
      DataTable dt = new DataTable("Convex Hull Volume over Generations");
      DataRow dtVolumeRow = new DataRow("Volume");
      DataRow dtNrPointsRow = new DataRow("Nr. of Points");
      dt.Rows.Add(dtVolumeRow);
      dt.Rows.Add(dtNrPointsRow);

      while (sols.Count != 0) {
        var input = sols.Select(x => ConvertPermutationToVertex(x)).ToArray();
        var convexHull = LPHull.Calculate(input);
        var volume = ConvexHullMeasures.CalculateVolume(convexHull);
        dtVolumeRow.Values.Add(volume);
        dtNrPointsRow.Values.Add(convexHull.Count);

        sols = solutionCache.GetSolutionsFromGeneration(++i);
      }

      results.Add(new Result("Convex Hull Measures", dt));
    }

    private double[] ConvertPermutationToVertex(RealVector p) {
      double[] vertex = p.Select(x => (double)x).ToArray();
      return vertex;
    }
  }
}
